
#include <stdio.h>
#include <mpi.h>

#include "typeinfo.h"
#include "allocation.h"
#include "init.h"
#include "stencil.h"
#include "stencnst.h"
#include "comm.h"

void calculate(
	const int nx,
	const int ny,
	const int nz,
	const int num_vars,
	const int percent_sum,
	const int num_spikes,
	const int num_tsteps,
	const int stencil,
	const int comm_method,
	const int send_protocol,
	const int check_diffusion,
	const int npx,
	const int npy,
	const int npz,
	const int profiling) {
	
	int rank;
	MPI_Comm_rank( MPI_COMM_WORLD, &rank );
	
	int size;
	MPI_Comm_size( MPI_COMM_WORLD, &size );
	
	int my_x, my_y, my_z;
	my_y = (rank % (npx * npy)) / npx;
	my_x = (rank % (npx * npy)) % npx;
	my_z = rank / (npx * npy);
	
	int neigh_south = (my_y == 0) ? -1 : rank - npx;
	int neigh_north = (my_y == npy - 1) ? -1 : rank + npx;
	int neigh_west  = (my_x == 0) ? -1 : rank - 1;
	int neigh_east  = (my_x == npx - 1) ? -1 : rank + 1;
	int neigh_back  = (my_z == 0) ? -1 : rank - (npx * npy);
	int neigh_front = (my_z == npz - 1) ? -1 : rank + (npy * npy);
	
	MPI_Barrier(MPI_COMM_WORLD);
	if(rank == 0) {
		printf("Launching stencil computation...\n");
	}
	MPI_Barrier(MPI_COMM_WORLD);
	
	SCALAR_TYPE****	data   = allocate_data(num_vars, nx, ny, nz);
	SCALAR_TYPE**** result = allocate_data(num_vars, nx, ny, nz);
	
	// zero out data grid
	init_grid(data, num_vars, nx, ny, nz);

	// put a "hot corner" in rank 0
	if(rank == 0) {
		int var_counter = 0;
		
		for(var_counter = 0; var_counter < num_vars; var_counter++) {
			data[var_counter][nz / 2][ny / 2][nx / 2] = 1;
		}
	}
	
	int ts_counter;
	int z_counter, y_counter, x_counter, v_counter;
	
	SCALAR_TYPE localsum = 0;
	SCALAR_TYPE globalsum = 0;
	
	// init the comms
	init_buffers(num_vars, nx, ny, nz, neigh_north,
		neigh_south, neigh_east, neigh_west, neigh_back, neigh_front);
	
	///////////////////////////////////////////////////////////////////////////////////
	
	for(ts_counter = 1; ts_counter <= num_tsteps; ts_counter++) {
		// zero out result grid
		init_grid(result, num_vars, nx, ny, nz);
	
		switch(stencil) {
			case STENCIL_3D27PT :
				stencil_3d27pt(data, result, num_vars, nx, ny, nz);
				break;
			
			case STENCIL_2D9PT :
				stencil_2d9pt(data, result, num_vars, nx, ny, nz);
				break;
				
			default:
				if(rank == 0) {
					fprintf(stderr, "! Error: unknown stencil %d type.\n",
						stencil);
					MPI_Abort(MPI_COMM_WORLD, -1);
				}
		}
		
		// exchange via MPI
		if(size > 1) {
			exchange(data, result, num_vars, nx, ny, nz, 
				neigh_north, neigh_south, 
				neigh_east, neigh_west, 
				neigh_back, neigh_front);
		}
		
		// copy results back
		copy_grid(result, data, num_vars, nx, ny, nz);
		
		if((ts_counter+1) % 20 == 0 || ts_counter == (num_tsteps - 1)) {
			// check the global value
			localsum = 0;
			for(v_counter = 0; v_counter < num_vars; v_counter++) {
				for(z_counter = 1; z_counter <= nz; z_counter++) {
					for(y_counter = 1; y_counter <= ny; y_counter++) {
						for(x_counter = 1; x_counter <= nx; x_counter++) {
							localsum += data[v_counter][z_counter][y_counter][x_counter];
						}
					}
				}
			}
		
			MPI_Allreduce(&localsum, &globalsum, 1, SCALAR_MPI_TYPE, MPI_SUM, MPI_COMM_WORLD);
	
			if(rank == 0) {
				printf("[Step: %6d] global value: %5.15f\n", (ts_counter+1), globalsum);
			}	
		}
	}
}
