
#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>

#include "calculate.h"
#include "stencnst.h"

#define SCALING_STRONG 1
#define SCALING_WEAK   2

int main(int argc, char* argv[]) {

	int error_check;
	error_check = MPI_Init( &argc, &argv );

	int rank = 0;
	int size = 0;
	MPI_Comm_rank( MPI_COMM_WORLD, &rank );
	MPI_Comm_size( MPI_COMM_WORLD, &size );

	if(rank == 0) {
		printf("miniGhost C\n");
		printf("----------------------------------------------------------------------\n");
	}

	/* set up defaults to be overridden by user input */
	int scaling = 2;
	int nx = 10;
	int ny = 10;
	int nz = 10;
	int num_vars = 5;
	int percent_sum = 0;
	int num_spikes = 1;
	int num_tsteps = 10;
	int stencil = STENCIL_3D27PT;
	int comm_method = 1;
	int send_protocol = 22;
	int check_diffusion = 1;
	int npx = 1;
	int npy = 1;
	int npz = 1;
	int profiling = 1;

	int i = 0;
	for(i = 1; i < argc; i++) {
		if( ! strcmp(argv[i], "--scaling") ) 
			scaling = atoi(argv[++i]);
		else if (! strcmp(argv[i], "--nx") ) 
			nx = atoi(argv[++i]);
		else if (! strcmp(argv[i], "--ny") ) 
			ny = atoi(argv[++i]);
		else if (! strcmp(argv[i], "--nz") ) 
			nz = atoi(argv[++i]);
		else if ( (! strcmp(argv[i], "--num_vars")) || (! strcmp(argv[i], "--nv")) )
			num_vars = atoi(argv[++i]);
		else if ( (! strcmp(argv[i], "--num_tsteps")) || (! strcmp(argv[i], "--ts")) )
			num_tsteps = atoi(argv[++i]);
		else if ( (! strcmp(argv[i], "--percent_sum")) ) 
			percent_sum = atoi(argv[++i]);
		else if ( (! strcmp(argv[i], "--num_spikes")) )
			num_spikes = atoi(argv[++i]);
		else if ( (! strcmp(argv[i], "--stencil")) ) 
			stencil = atoi(argv[++i]);
		else if ( (! strcmp(argv[i], "--comm_method")) )
			comm_method = atoi(argv[++i]);
		else if ( (! strcmp( argv[i], "--send_protocol" )) )
			send_protocol = atoi(argv[++i]);
		else if ( (! strcmp( argv[i], "--check_diffusion" )) ) 
			check_diffusion = atoi(argv[i++]);
		else if ( (! strcmp( argv[i], "--npx")) || (! strcmp(argv[i], "--px")) )
			npx = atoi(argv[++i]);
		else if ( (! strcmp( argv[i], "--npy")) || (! strcmp(argv[i], "--py")) )
			npy = atoi(argv[++i]);
		else if ( (! strcmp( argv[i], "--npz")) || (! strcmp(argv[i], "--pz")) ) 
			npz = atoi(argv[++i]);
		else if ( (! strcmp(argv[i], "--profiling")) )
			profiling = atoi(argv[++i]);
		else {
			if(rank == 0) {
				fprintf(stderr, "! Error: unknown parameter: %s (index=%d)\n",
					argv[i], i);
				MPI_Abort(MPI_COMM_WORLD, -1);
			}
		}
	}
	
	// ----- perform initial checks -----
	if(npx * npy * npz != size) {
		if(rank == 0) {
			fprintf(stderr, "! Error: processors (%d * %d * %d) != %d\n", 
				npx, npy, npz, size);
			MPI_Abort(MPI_COMM_WORLD, -1);	
		}
	}
	
	// ----- compute problem logic -----
	
	if( scaling == SCALING_WEAK ) {
		nx = nx * npx;
		ny = ny * npy;
		nz = nz * npz;
	} else if ( scaling == SCALING_STRONG ) {
		int remainder = nx % npx;
		nx = (rank < remainder) ? (nx / npx) + 1 : (nx / npx);
		
		remainder = ny % npy;
		nx = (rank < remainder) ? (ny / npy) + 1 : (ny / npy);
		
		remainder = nz % npz;
		nz = (rank < remainder) ? (nz / npz) + 1 : (nz / npz);
	} else {
		if(rank == 0) {
			fprintf(stderr, "! Error: unknown scaling type: %d\n", scaling);
			MPI_Abort(MPI_COMM_WORLD, -1);
		}
	}
	
	calculate(nx, ny, nz, num_vars, percent_sum,
		num_spikes, num_tsteps, stencil, comm_method,
		send_protocol, check_diffusion, npx, npy, npz,
		profiling);
		
	MPI_Barrier(MPI_COMM_WORLD);
	
	if(rank == 0) {
		printf("Calculation completed.\n");
	}
	
	MPI_Finalize();
}